(* file kernel/ocaml/mli/numerix.mli: Objective-Caml extensible integer definitions
 +-----------------------------------------------------------------------+
 |  Copyright 2005-2006, Michel Quercia (michel.quercia@prepas.org)      |
 |                                                                       |
 |  This file is part of Numerix. Numerix is free software; you can      |
 |  redistribute it and/or modify it under the terms of the GNU Lesser   |
 |  General Public License as published by the Free Software Foundation; |
 |  either version 2.1 of the License, or (at your option) any later     |
 |  version.                                                             |
 |                                                                       |
 |  The Numerix Library is distributed in the hope that it will be       |
 |  useful, but WITHOUT ANY WARRANTY; without even the implied warranty  |
 |  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  |
 |  Lesser General Public License for more details.                      |
 |                                                                       |
 |  You should have received a copy of the GNU Lesser General Public     |
 |  License along with the GNU MP Library; see the file COPYING. If not, |
 |  write to the Free Software Foundation, Inc., 59 Temple Place -       |
 |  Suite 330, Boston, MA 02111-1307, USA.                               |
 +-----------------------------------------------------------------------+
 |                                                                       |
 |                           Dfinitions                                 |
 |                                                                       |
 +-----------------------------------------------------------------------*)

                       (* +-------------------------+
                          |  Description abstraite  |
                          +-------------------------+ *)

(* mode d'arrondi *)
type round_mode = Floor | Nearest_up | Ceil | Nearest_down

(* rsultat ternaire *)
type tristate = False | Unknown | True

module type Int_type = sig

  type t                     (* entier         *)
  type tref                  (* entier mutable *)

  val name : unit -> string  (* nom du module  *)

  val zero : t               (* the number 0   *)
  val one  : t               (* the number 1   *)

  (* rfrence      ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val make_ref    :                               t ->               tref   
  val copy_in     :               tref ->         t ->               unit   
  val copy_out    :               tref ->                            t      
  val look        :               tref ->                            t      
  
  (* addition       ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val add         :                               t -> t   ->        t      
  val add_1       :                               t -> int ->        t      
  val add_in      :               tref ->         t -> t   ->        unit   
  val add_1_in    :               tref ->         t -> int ->        unit   
  
  (* soustraction   ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val sub         :                               t -> t   ->        t      
  val sub_1       :                               t -> int ->        t      
  val sub_in      :               tref ->         t -> t   ->        unit   
  val sub_1_in    :               tref ->         t -> int ->        unit   
  
  (* multiplication ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val mul         :                               t -> t   ->        t      
  val mul_1       :                               t -> int ->        t      
  val mul_in      :               tref ->         t -> t   ->        unit   
  val mul_1_in    :               tref ->         t -> int ->        unit   
  
  (* division       ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val quomod      :                               t -> t   ->        t*t    
  val quo         :                               t -> t   ->        t      
  val modulo      :                               t -> t   ->        t      
  val gquomod     : round_mode ->                 t -> t   ->        t*t    
  val gquo        : round_mode ->                 t -> t   ->        t      
  val gmod        : round_mode ->                 t -> t   ->        t      
  
  val quomod_in   :               tref -> tref -> t -> t   ->        unit   
  val quo_in      :               tref ->         t -> t   ->        unit   
  val mod_in      :                       tref -> t -> t   ->        unit   
  val gquomod_in  : round_mode -> tref -> tref -> t -> t   ->        unit   
  val gquo_in     : round_mode -> tref ->         t -> t   ->        unit   
  val gmod_in     : round_mode ->         tref -> t -> t   ->        unit   
  
  val quomod_1    :                               t -> int ->        t*int  
  val quo_1       :                               t -> int ->        t      
  val mod_1       :                               t -> int ->        int    
  val gquomod_1   : round_mode ->                 t -> int ->        t*int  
  val gquo_1      : round_mode ->                 t -> int ->        t      
  val gmod_1      : round_mode ->                 t -> int ->        int    
  
  val quomod_1_in :               tref ->         t -> int ->        int    
  val quo_1_in    :               tref ->         t -> int ->        unit   
  val gquomod_1_in: round_mode -> tref ->         t -> int ->        int    
  val gquo_1_in   : round_mode -> tref ->         t -> int ->        unit   
  
  (* valeur absolue ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val abs         :                               t ->               t      
  val abs_in      :               tref ->         t ->               unit   
  
  (* oppos         ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val neg         :                               t ->               t      
  val neg_in      :               tref ->         t ->               unit   
  
  (* puissance p-me----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val sqr         :                               t ->               t      
  val pow         :                               t -> int ->        t      
  val pow_1       :                             int -> int ->        t      
  val powmod      :                               t -> t   -> t ->   t      
  val gpowmod     : round_mode ->                 t -> t   -> t ->   t      
  val sqr_in      :               tref ->         t ->               unit   
  val pow_in      :               tref ->         t -> int ->        unit   
  val pow_1_in    :               tref ->       int -> int ->        unit   
  val powmod_in   :               tref ->         t -> t   -> t ->   unit   
  val gpowmod_in  : round_mode -> tref ->         t -> t   -> t ->   unit   
  
  (* racine p-me    ----------------------------------------------------- *)
  (*                 mode          r       s       a    b      c      res  *)
  val sqrt        :                               t ->               t      
  val root        :                               t -> int ->        t      
  val gsqrt       : round_mode ->                 t ->               t      
  val groot       : round_mode ->                 t -> int ->        t      
  val sqrt_in     :               tref ->         t ->               unit   
  val root_in     :               tref ->         t -> int ->        unit   
  val gsqrt_in    : round_mode -> tref ->         t ->               unit   
  val groot_in    : round_mode -> tref ->         t -> int ->        unit   
  
  (* factorielle    ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val fact        :                             int ->               t      
  val fact_in     :               tref ->       int ->               unit   
  
  (* pgcd           ----------------------------------------------------- *)
  (*                  d     u     v     p    q    a    b      c      res  *)
  val gcd         :                               t -> t   ->        t      
  val gcd_ex      :                               t -> t   ->        t*t*t  
  val cfrac       :                               t -> t   ->     t*t*t*t*t 
  val gcd_in      : tref->                        t -> t   ->        unit   
  val gcd_ex_in   : tref->tref->tref->            t -> t   ->        unit   
  val cfrac_in    : tref->tref->tref->tref->tref->t -> t   ->        unit   
  
  (* primalit      ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val isprime     :                               t ->               tristate
  val isprime_1   :                             int ->               tristate

  (* comparaison    ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val sgn         :                               t ->               int    
  val cmp         :                               t -> t   ->        int    
  val cmp_1       :                               t -> int ->        int    
  val eq          :                               t -> t   ->        bool   
  val eq_1        :                               t -> int ->        bool   
  val neq         :                               t -> t   ->        bool   
  val neq_1       :                               t -> int ->        bool   
  val inf         :                               t -> t   ->        bool   
  val inf_1       :                               t -> int ->        bool   
  val infeq       :                               t -> t   ->        bool   
  val infeq_1     :                               t -> int ->        bool   
  val sup         :                               t -> t   ->        bool   
  val sup_1       :                               t -> int ->        bool   
  val supeq       :                               t -> t   ->        bool   
  val supeq_1     :                               t -> int ->        bool   
  
  (* conversion     ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val of_int      :                             int ->               t      
  val of_string   :                          string ->               t      
  val of_int_in   :               tref ->       int ->               unit   
  val of_string_in:               tref ->    string ->               unit   
  val int_of      :                               t ->               int    
  val string_of   :                               t ->               string 
  val bstring_of  :                               t ->               string 
  val hstring_of  :                               t ->               string 
  val ostring_of  :                               t ->               string 
  
  (* nombre alatoire --------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val nrandom     :                             int->                t      
  val zrandom     :                             int->                t      
  val nrandom1    :                             int->                t      
  val zrandom1    :                             int->                t      
  val nrandom_in  :               tref ->       int->                unit   
  val zrandom_in  :               tref ->       int->                unit   
  val nrandom1_in :               tref ->       int->                unit   
  val zrandom1_in :               tref ->       int->                unit   
  val random_init :                             int->                unit   
  
  (* reprsentation binaire --------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val nbits       :                               t ->               int    
  val lowbits     :                               t ->               int    
  val highbits    :                               t ->               int    
  val nth_word    :                               t -> int ->        int    
  val nth_bit     :                               t -> int ->        bool   
  
  (* dcalage       ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val shl         :                               t -> int ->        t      
  val shr         :                               t -> int ->        t      
  val split       :                               t -> int ->        t*t    
  val join        :                               t -> t   -> int -> t      
  val shl_in      :               tref ->         t -> int ->        unit   
  val shr_in      :               tref ->         t -> int ->        unit   
  val split_in    :               tref -> tref -> t -> int ->        unit   
  val join_in     :               tref ->         t -> t   -> int -> unit   
  
  (* affichage      ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val toplevel_print   :                               t ->               unit
  val toplevel_print_tref:             tref ->                            unit
  
  (* exceptions *)
  exception Error of string

end

                          (* +-------------------+
                             |  notation infixe  |
                             +-------------------+ *)

module Infixes(E : Int_type) : sig
  open E

  val (  ++  ) :   t -> t   -> t      (* add      *)
  val (  --  ) :   t -> t   -> t      (* sub      *)
  val (  **  ) :   t -> t   -> t      (* mul      *)
  val (  //  ) :   t -> t   -> t      (* quo      *)
  val (  %%  ) :   t -> t   -> t      (* modulo   *)
  val (  /%  ) :   t -> t   -> t*t    (* quomod   *)
  val (  <<  ) :   t -> int -> t      (* shl      *)
  val (  >>  ) :   t -> int -> t      (* shr      *)
  val (  ^^  ) :   t -> int -> t      (* pow      *)
                 
  val (  +=  ) : tref -> t -> unit    (* add_in   *)
  val (  -=  ) : tref -> t -> unit    (* sub_in   *)
  val (  *=  ) : tref -> t -> unit    (* mul_in   *)
  val (  /=  ) : tref -> t -> unit    (* quo_in   *)
  val (  %=  ) : tref -> t -> unit    (* mod_in  *)
                
  val (  +.  ) :   t -> int -> t      (* add_1    *)
  val (  -.  ) :   t -> int -> t      (* sub_1    *)
  val (  *.  ) :   t -> int -> t      (* mul_1    *)
  val (  /.  ) :   t -> int -> t      (* quo_1    *)
  val (  %.  ) :   t -> int -> int    (* mod_1    *)
  val (  /%. ) :   t -> int -> t*int  (* quomod_1 *)
  val (  ^.  ) : int -> int -> t      (* pow_1    *)
                 
  val (  +=. ) : tref -> int -> unit  (* add_1_in *)
  val (  -=. ) : tref -> int -> unit  (* sub_1_in *)
  val (  *=. ) : tref -> int -> unit  (* mul_1_in *)
  val (  /=. ) : tref -> int -> unit  (* quo_1_in *)

  val (   =. ) :   t -> int -> bool   (* eq_1     *)
  val (  <>. ) :   t -> int -> bool   (* neq_1    *)
  val (   <. ) :   t -> int -> bool   (* inf_1    *)
  val (  <=. ) :   t -> int -> bool   (* infeq_1  *)
  val (   >. ) :   t -> int -> bool   (* sup_1    *)
  val (  >=. ) :   t -> int -> bool   (* supeq_1  *)
                 
  val (  ~~  ) : tref -> t            (* look     *)

end

                           (* +-----------------+
                              |  Chronomtrage  |
                              +-----------------+ *)

external chrono : string -> unit = "chrono"


                           (* +-----------------+
                              |  module Clong   |
                              +-----------------+ *)

module Clong : sig

  type t
  type tref
  val name : unit -> string
  val zero : t
  val one  : t

  external make_ref    :                               t ->               tref   = "cx_make_ref"
  external copy_in     :               tref ->         t ->               unit   = "cx_copy"
  external copy_out    :               tref ->                            t      = "cx_copy_out"
  external look        :               tref ->                            t      = "cx_look"
  external add         :                               t -> t   ->        t      = "cx_f_add"
  external add_1       :                               t -> int ->        t      = "cx_f_add_1"
  external add_in      :               tref ->         t -> t   ->        unit   = "cx_add"
  external add_1_in    :               tref ->         t -> int ->        unit   = "cx_add_1"
  external sub         :                               t -> t   ->        t      = "cx_f_sub"
  external sub_1       :                               t -> int ->        t      = "cx_f_sub_1"
  external sub_in      :               tref ->         t -> t   ->        unit   = "cx_sub"
  external sub_1_in    :               tref ->         t -> int ->        unit   = "cx_sub_1"
  external mul         :                               t -> t   ->        t      = "cx_f_mul"
  external mul_1       :                               t -> int ->        t      = "cx_f_mul_1"
  external mul_in      :               tref ->         t -> t   ->        unit   = "cx_mul"
  external mul_1_in    :               tref ->         t -> int ->        unit   = "cx_mul_1"
  external quomod      :                               t -> t   ->        t*t    = "cx_f_quomod"
  external quo         :                               t -> t   ->        t      = "cx_f_quo"
  external modulo      :                               t -> t   ->        t      = "cx_f_mod"
  external gquomod     : round_mode ->                 t -> t   ->        t*t    = "cx_f_gquomod"
  external gquo        : round_mode ->                 t -> t   ->        t      = "cx_f_gquo"
  external gmod        : round_mode ->                 t -> t   ->        t      = "cx_f_gmod"
  external quomod_in   :               tref -> tref -> t -> t   ->        unit   = "cx_quomod"
  external quo_in      :               tref ->         t -> t   ->        unit   = "cx_quo"
  external mod_in      :                       tref -> t -> t   ->        unit   = "cx_mod"
  external gquomod_in  : round_mode -> tref -> tref -> t -> t   ->        unit   = "cx_gquomod"
  external gquo_in     : round_mode -> tref ->         t -> t   ->        unit   = "cx_gquo"
  external gmod_in     : round_mode ->         tref -> t -> t   ->        unit   = "cx_gmod"
  external quomod_1    :                               t -> int ->        t*int  = "cx_f_quomod_1"
  external quo_1       :                               t -> int ->        t      = "cx_f_quo_1"
  external mod_1       :                               t -> int ->        int    = "cx_f_mod_1"
  external gquomod_1   : round_mode ->                 t -> int ->        t*int  = "cx_f_gquomod_1"
  external gquo_1      : round_mode ->                 t -> int ->        t      = "cx_f_gquo_1"
  external gmod_1      : round_mode ->                 t -> int ->        int    = "cx_f_gmod_1"
  external quomod_1_in :               tref ->         t -> int ->        int    = "cx_quomod_1"
  external quo_1_in    :               tref ->         t -> int ->        unit   = "cx_quo_1"
  external gquomod_1_in: round_mode -> tref ->         t -> int ->        int    = "cx_gquomod_1"
  external gquo_1_in   : round_mode -> tref ->         t -> int ->        unit   = "cx_gquo_1"
  external abs         :                               t ->               t      = "cx_f_abs"
  external abs_in      :               tref ->         t ->               unit   = "cx_abs"
  external neg         :                               t ->               t      = "cx_f_neg"
  external neg_in      :               tref ->         t ->               unit   = "cx_neg"
  external sqr         :                               t ->               t      = "cx_f_sqr"
  external pow         :                               t -> int ->        t      = "cx_f_pow"
  external pow_1       :                             int -> int ->        t      = "cx_f_pow_1"
  external powmod      :                               t -> t   -> t ->   t      = "cx_f_powmod"
  external gpowmod     : round_mode ->                 t -> t   -> t ->   t      = "cx_f_gpowmod"
  external sqr_in      :               tref ->         t ->               unit   = "cx_sqr"
  external pow_in      :               tref ->         t -> int ->        unit   = "cx_pow"
  external pow_1_in    :               tref ->       int -> int ->        unit   = "cx_pow_1"
  external powmod_in   :               tref ->         t -> t   -> t ->   unit   = "cx_powmod"
  external gpowmod_in  : round_mode -> tref ->         t -> t   -> t ->   unit   = "cx_gpowmod"
  external sqrt        :                               t ->               t      = "cx_f_sqrt"
  external root        :                               t -> int ->        t      = "cx_f_root"
  external gsqrt       : round_mode ->                 t ->               t      = "cx_f_gsqrt"
  external groot       : round_mode ->                 t -> int ->        t      = "cx_f_groot"
  external sqrt_in     :               tref ->         t ->               unit   = "cx_sqrt"
  external root_in     :               tref ->         t -> int ->        unit   = "cx_root"
  external gsqrt_in    : round_mode -> tref ->         t ->               unit   = "cx_gsqrt"
  external groot_in    : round_mode -> tref ->         t -> int ->        unit   = "cx_groot"
  external fact        :                             int ->               t      = "cx_f_fact"
  external fact_in     :               tref ->       int ->               unit   = "cx_fact"
  external gcd         :                               t -> t   ->        t      = "cx_f_gcd"
  external gcd_ex      :                               t -> t   ->        t*t*t  = "cx_f_gcd_ex"
  external cfrac       :                               t -> t   ->     t*t*t*t*t = "cx_f_cfrac"
  external gcd_in      : tref->                        t -> t   ->        unit   = "cx_gcd"
  external gcd_ex_in   : tref->tref->tref->            t -> t   ->        unit   = "cx_gcd_ex"
  external cfrac_in    : tref->tref->tref->tref->tref->t -> t   ->        unit   = "cx_cfrac_bytecode" "cx_cfrac"
  external isprime     :                               t ->           tristate   = "cx_isprime"
  external isprime_1   :                             int ->           tristate   = "cx_isprime_1"
  external sgn         :                               t ->               int    = "cx_sgn"
  external cmp         :                               t -> t   ->        int    = "cx_cmp"
  external cmp_1       :                               t -> int ->        int    = "cx_cmp_1"
  external eq          :                               t -> t   ->        bool   = "cx_eq"
  external eq_1        :                               t -> int ->        bool   = "cx_eq_1"
  external neq         :                               t -> t   ->        bool   = "cx_neq"
  external neq_1       :                               t -> int ->        bool   = "cx_neq_1"
  external inf         :                               t -> t   ->        bool   = "cx_inf"
  external inf_1       :                               t -> int ->        bool   = "cx_inf_1"
  external infeq       :                               t -> t   ->        bool   = "cx_infeq"
  external infeq_1     :                               t -> int ->        bool   = "cx_infeq_1"
  external sup         :                               t -> t   ->        bool   = "cx_sup"
  external sup_1       :                               t -> int ->        bool   = "cx_sup_1"
  external supeq       :                               t -> t   ->        bool   = "cx_supeq"
  external supeq_1     :                               t -> int ->        bool   = "cx_supeq_1"
  external of_int      :                             int ->               t      = "cx_of_int"
  external of_string   :                          string ->               t      = "cx_of_string"
  external of_int_in   :               tref ->       int ->               unit   = "cx_copy_int"
  external of_string_in:               tref ->    string ->               unit   = "cx_copy_string"
  external int_of      :                               t ->               int    = "cx_int_of"
  external string_of   :                               t ->               string = "cx_string_of"
  external bstring_of  :                               t ->               string = "cx_bstring_of"
  external hstring_of  :                               t ->               string = "cx_hstring_of"
  external ostring_of  :                               t ->               string = "cx_ostring_of"
  external nrandom     :                             int->                t      = "cx_f_nrandom"
  external zrandom     :                             int->                t      = "cx_f_zrandom"
  external nrandom1    :                             int->                t      = "cx_f_nrandom1"
  external zrandom1    :                             int->                t      = "cx_f_zrandom1"
  external nrandom_in  :               tref ->       int->                unit   = "cx_nrandom"
  external zrandom_in  :               tref ->       int->                unit   = "cx_zrandom"
  external nrandom1_in :               tref ->       int->                unit   = "cx_nrandom1"
  external zrandom1_in :               tref ->       int->                unit   = "cx_zrandom1"
  val      random_init :                             int->                unit
  external nbits       :                               t ->               int    = "cx_nbits"
  external lowbits     :                               t ->               int    = "cx_lowbits"
  external highbits    :                               t ->               int    = "cx_highbits"
  external nth_word    :                               t -> int ->        int    = "cx_nth_word"
  external nth_bit     :                               t -> int ->        bool   = "cx_nth_bit"
  external shl         :                               t -> int ->        t      = "cx_f_shl"
  external shr         :                               t -> int ->        t      = "cx_f_shr"
  external split       :                               t -> int ->        t*t    = "cx_f_split"
  external join        :                               t -> t   -> int -> t      = "cx_f_join"
  external shl_in      :               tref ->         t -> int ->        unit   = "cx_shl"
  external shr_in      :               tref ->         t -> int ->        unit   = "cx_shr"
  external split_in    :               tref -> tref -> t -> int ->        unit   = "cx_split"
  external join_in     :               tref ->         t -> t   -> int -> unit   = "cx_join"
  external dump        :                               t ->               unit   = "cx_dump"
  val toplevel_print   :                               t ->               unit
  val toplevel_print_tref:             tref ->                            unit
  exception Error of string

end

                           (* +-----------------+
                              |  module Slong   |
                              +-----------------+ *)

module Slong : sig

  type t
  type tref
  val name : unit -> string
  val zero : t
  val one  : t

  external make_ref    :                               t ->               tref   = "sx_make_ref"
  external copy_in     :               tref ->         t ->               unit   = "sx_copy"
  external copy_out    :               tref ->                            t      = "sx_copy_out"
  external look        :               tref ->                            t      = "sx_look"
  external add         :                               t -> t   ->        t      = "sx_f_add"
  external add_1       :                               t -> int ->        t      = "sx_f_add_1"
  external add_in      :               tref ->         t -> t   ->        unit   = "sx_add"
  external add_1_in    :               tref ->         t -> int ->        unit   = "sx_add_1"
  external sub         :                               t -> t   ->        t      = "sx_f_sub"
  external sub_1       :                               t -> int ->        t      = "sx_f_sub_1"
  external sub_in      :               tref ->         t -> t   ->        unit   = "sx_sub"
  external sub_1_in    :               tref ->         t -> int ->        unit   = "sx_sub_1"
  external mul         :                               t -> t   ->        t      = "sx_f_mul"
  external mul_1       :                               t -> int ->        t      = "sx_f_mul_1"
  external mul_in      :               tref ->         t -> t   ->        unit   = "sx_mul"
  external mul_1_in    :               tref ->         t -> int ->        unit   = "sx_mul_1"
  external quomod      :                               t -> t   ->        t*t    = "sx_f_quomod"
  external quo         :                               t -> t   ->        t      = "sx_f_quo"
  external modulo      :                               t -> t   ->        t      = "sx_f_mod"
  external gquomod     : round_mode ->                 t -> t   ->        t*t    = "sx_f_gquomod"
  external gquo        : round_mode ->                 t -> t   ->        t      = "sx_f_gquo"
  external gmod        : round_mode ->                 t -> t   ->        t      = "sx_f_gmod"
  external quomod_in   :               tref -> tref -> t -> t   ->        unit   = "sx_quomod"
  external quo_in      :               tref ->         t -> t   ->        unit   = "sx_quo"
  external mod_in      :                       tref -> t -> t   ->        unit   = "sx_mod"
  external gquomod_in  : round_mode -> tref -> tref -> t -> t   ->        unit   = "sx_gquomod"
  external gquo_in     : round_mode -> tref ->         t -> t   ->        unit   = "sx_gquo"
  external gmod_in     : round_mode ->         tref -> t -> t   ->        unit   = "sx_gmod"
  external quomod_1    :                               t -> int ->        t*int  = "sx_f_quomod_1"
  external quo_1       :                               t -> int ->        t      = "sx_f_quo_1"
  external mod_1       :                               t -> int ->        int    = "sx_f_mod_1"
  external gquomod_1   : round_mode ->                 t -> int ->        t*int  = "sx_f_gquomod_1"
  external gquo_1      : round_mode ->                 t -> int ->        t      = "sx_f_gquo_1"
  external gmod_1      : round_mode ->                 t -> int ->        int    = "sx_f_gmod_1"
  external quomod_1_in :               tref ->         t -> int ->        int    = "sx_quomod_1"
  external quo_1_in    :               tref ->         t -> int ->        unit   = "sx_quo_1"
  external gquomod_1_in: round_mode -> tref ->         t -> int ->        int    = "sx_gquomod_1"
  external gquo_1_in   : round_mode -> tref ->         t -> int ->        unit   = "sx_gquo_1"
  external abs         :                               t ->               t      = "sx_f_abs"
  external abs_in      :               tref ->         t ->               unit   = "sx_abs"
  external neg         :                               t ->               t      = "sx_f_neg"
  external neg_in      :               tref ->         t ->               unit   = "sx_neg"
  external sqr         :                               t ->               t      = "sx_f_sqr"
  external pow         :                               t -> int ->        t      = "sx_f_pow"
  external pow_1       :                             int -> int ->        t      = "sx_f_pow_1"
  external powmod      :                               t -> t   -> t ->   t      = "sx_f_powmod"
  external gpowmod     : round_mode ->                 t -> t   -> t ->   t      = "sx_f_gpowmod"
  external sqr_in      :               tref ->         t ->               unit   = "sx_sqr"
  external pow_in      :               tref ->         t -> int ->        unit   = "sx_pow"
  external pow_1_in    :               tref ->       int -> int ->        unit   = "sx_pow_1"
  external powmod_in   :               tref ->         t -> t   -> t ->   unit   = "sx_powmod"
  external gpowmod_in  : round_mode -> tref ->         t -> t   -> t ->   unit   = "sx_gpowmod"
  external sqrt        :                               t ->               t      = "sx_f_sqrt"
  external root        :                               t -> int ->        t      = "sx_f_root"
  external gsqrt       : round_mode ->                 t ->               t      = "sx_f_gsqrt"
  external groot       : round_mode ->                 t -> int ->        t      = "sx_f_groot"
  external sqrt_in     :               tref ->         t ->               unit   = "sx_sqrt"
  external root_in     :               tref ->         t -> int ->        unit   = "sx_root"
  external gsqrt_in    : round_mode -> tref ->         t ->               unit   = "sx_gsqrt"
  external groot_in    : round_mode -> tref ->         t -> int ->        unit   = "sx_groot"
  external fact        :                             int ->               t      = "sx_f_fact"
  external fact_in     :               tref ->       int ->               unit   = "sx_fact"
  external gcd         :                               t -> t   ->        t      = "sx_f_gcd"
  external gcd_ex      :                               t -> t   ->        t*t*t  = "sx_f_gcd_ex"
  external cfrac       :                               t -> t   ->     t*t*t*t*t = "sx_f_cfrac"
  external gcd_in      : tref->                        t -> t   ->        unit   = "sx_gcd"
  external gcd_ex_in   : tref->tref->tref->            t -> t   ->        unit   = "sx_gcd_ex"
  external cfrac_in    : tref->tref->tref->tref->tref->t -> t   ->        unit   = "sx_cfrac_bytecode" "sx_cfrac"
  external isprime     :                               t ->           tristate   = "sx_isprime"
  external isprime_1   :                             int ->           tristate   = "sx_isprime_1"
  external sgn         :                               t ->               int    = "sx_sgn"
  external cmp         :                               t -> t   ->        int    = "sx_cmp"
  external cmp_1       :                               t -> int ->        int    = "sx_cmp_1"
  external eq          :                               t -> t   ->        bool   = "sx_eq"
  external eq_1        :                               t -> int ->        bool   = "sx_eq_1"
  external neq         :                               t -> t   ->        bool   = "sx_neq"
  external neq_1       :                               t -> int ->        bool   = "sx_neq_1"
  external inf         :                               t -> t   ->        bool   = "sx_inf"
  external inf_1       :                               t -> int ->        bool   = "sx_inf_1"
  external infeq       :                               t -> t   ->        bool   = "sx_infeq"
  external infeq_1     :                               t -> int ->        bool   = "sx_infeq_1"
  external sup         :                               t -> t   ->        bool   = "sx_sup"
  external sup_1       :                               t -> int ->        bool   = "sx_sup_1"
  external supeq       :                               t -> t   ->        bool   = "sx_supeq"
  external supeq_1     :                               t -> int ->        bool   = "sx_supeq_1"
  external of_int      :                             int ->               t      = "sx_of_int"
  external of_string   :                          string ->               t      = "sx_of_string"
  external of_int_in   :               tref ->       int ->               unit   = "sx_copy_int"
  external of_string_in:               tref ->    string ->               unit   = "sx_copy_string"
  external int_of      :                               t ->               int    = "sx_int_of"
  external string_of   :                               t ->               string = "sx_string_of"
  external bstring_of  :                               t ->               string = "sx_bstring_of"
  external hstring_of  :                               t ->               string = "sx_hstring_of"
  external ostring_of  :                               t ->               string = "sx_ostring_of"
  external nrandom     :                             int->                t      = "sx_f_nrandom"
  external zrandom     :                             int->                t      = "sx_f_zrandom"
  external nrandom1    :                             int->                t      = "sx_f_nrandom1"
  external zrandom1    :                             int->                t      = "sx_f_zrandom1"
  external nrandom_in  :               tref ->       int->                unit   = "sx_nrandom"
  external zrandom_in  :               tref ->       int->                unit   = "sx_zrandom"
  external nrandom1_in :               tref ->       int->                unit   = "sx_nrandom1"
  external zrandom1_in :               tref ->       int->                unit   = "sx_zrandom1"
  val      random_init :                             int->                unit
  external nbits       :                               t ->               int    = "sx_nbits"
  external lowbits     :                               t ->               int    = "sx_lowbits"
  external highbits    :                               t ->               int    = "sx_highbits"
  external nth_word    :                               t -> int ->        int    = "sx_nth_word"
  external nth_bit     :                               t -> int ->        bool   = "sx_nth_bit"
  external shl         :                               t -> int ->        t      = "sx_f_shl"
  external shr         :                               t -> int ->        t      = "sx_f_shr"
  external split       :                               t -> int ->        t*t    = "sx_f_split"
  external join        :                               t -> t   -> int -> t      = "sx_f_join"
  external shl_in      :               tref ->         t -> int ->        unit   = "sx_shl"
  external shr_in      :               tref ->         t -> int ->        unit   = "sx_shr"
  external split_in    :               tref -> tref -> t -> int ->        unit   = "sx_split"
  external join_in     :               tref ->         t -> t   -> int -> unit   = "sx_join"
  external dump        :                               t ->               unit   = "sx_dump"
  val toplevel_print   :                               t ->               unit
  val toplevel_print_tref:             tref ->                            unit
  exception Error of string

end

                           (* +----------------+
                              |   module Big   |
                              +----------------+ *)

module Big : Int_type


                           (* +-----------------+
                              |  module Gmp   |
                              +-----------------+ *)

module Gmp : sig

  type t
  type tref
  val name : unit -> string
  val zero : t
  val one  : t

  external make_ref    :                               t ->               tref   = "gx_make_ref"
  external copy_in     :               tref ->         t ->               unit   = "gx_copy"
  external copy_out    :               tref ->                            t      = "gx_copy_out"
  external look        :               tref ->                            t      = "gx_look"
  external add         :                               t -> t   ->        t      = "gx_f_add"
  external add_1       :                               t -> int ->        t      = "gx_f_add_1"
  external add_in      :               tref ->         t -> t   ->        unit   = "gx_add"
  external add_1_in    :               tref ->         t -> int ->        unit   = "gx_add_1"
  external sub         :                               t -> t   ->        t      = "gx_f_sub"
  external sub_1       :                               t -> int ->        t      = "gx_f_sub_1"
  external sub_in      :               tref ->         t -> t   ->        unit   = "gx_sub"
  external sub_1_in    :               tref ->         t -> int ->        unit   = "gx_sub_1"
  external mul         :                               t -> t   ->        t      = "gx_f_mul"
  external mul_1       :                               t -> int ->        t      = "gx_f_mul_1"
  external mul_in      :               tref ->         t -> t   ->        unit   = "gx_mul"
  external mul_1_in    :               tref ->         t -> int ->        unit   = "gx_mul_1"
  external quomod      :                               t -> t   ->        t*t    = "gx_f_quomod"
  external quo         :                               t -> t   ->        t      = "gx_f_quo"
  external modulo      :                               t -> t   ->        t      = "gx_f_mod"
  external gquomod     : round_mode ->                 t -> t   ->        t*t    = "gx_f_gquomod"
  external gquo        : round_mode ->                 t -> t   ->        t      = "gx_f_gquo"
  external gmod        : round_mode ->                 t -> t   ->        t      = "gx_f_gmod"
  external quomod_in   :               tref -> tref -> t -> t   ->        unit   = "gx_quomod"
  external quo_in      :               tref ->         t -> t   ->        unit   = "gx_quo"
  external mod_in      :                       tref -> t -> t   ->        unit   = "gx_mod"
  external gquomod_in  : round_mode -> tref -> tref -> t -> t   ->        unit   = "gx_gquomod"
  external gquo_in     : round_mode -> tref ->         t -> t   ->        unit   = "gx_gquo"
  external gmod_in     : round_mode ->         tref -> t -> t   ->        unit   = "gx_gmod"
  external quomod_1    :                               t -> int ->        t*int  = "gx_f_quomod_1"
  external quo_1       :                               t -> int ->        t      = "gx_f_quo_1"
  external mod_1       :                               t -> int ->        int    = "gx_f_mod_1"
  external gquomod_1   : round_mode ->                 t -> int ->        t*int  = "gx_f_gquomod_1"
  external gquo_1      : round_mode ->                 t -> int ->        t      = "gx_f_gquo_1"
  external gmod_1      : round_mode ->                 t -> int ->        int    = "gx_f_gmod_1"
  external quomod_1_in :               tref ->         t -> int ->        int    = "gx_quomod_1"
  external quo_1_in    :               tref ->         t -> int ->        unit   = "gx_quo_1"
  external gquomod_1_in: round_mode -> tref ->         t -> int ->        int    = "gx_gquomod_1"
  external gquo_1_in   : round_mode -> tref ->         t -> int ->        unit   = "gx_gquo_1"
  external abs         :                               t ->               t      = "gx_f_abs"
  external abs_in      :               tref ->         t ->               unit   = "gx_abs"
  external neg         :                               t ->               t      = "gx_f_neg"
  external neg_in      :               tref ->         t ->               unit   = "gx_neg"
  external sqr         :                               t ->               t      = "gx_f_sqr"
  external pow         :                               t -> int ->        t      = "gx_f_pow"
  external pow_1       :                             int -> int ->        t      = "gx_f_pow_1"
  external powmod      :                               t -> t   -> t ->   t      = "gx_f_powmod"
  external gpowmod     : round_mode ->                 t -> t   -> t ->   t      = "gx_f_gpowmod"
  external sqr_in      :               tref ->         t ->               unit   = "gx_sqr"
  external pow_in      :               tref ->         t -> int ->        unit   = "gx_pow"
  external pow_1_in    :               tref ->       int -> int ->        unit   = "gx_pow_1"
  external powmod_in   :               tref ->         t -> t   -> t ->   unit   = "gx_powmod"
  external gpowmod_in  : round_mode -> tref ->         t -> t   -> t ->   unit   = "gx_gpowmod"
  external sqrt        :                               t ->               t      = "gx_f_sqrt"
  external root        :                               t -> int ->        t      = "gx_f_root"
  external gsqrt       : round_mode ->                 t ->               t      = "gx_f_gsqrt"
  external groot       : round_mode ->                 t -> int ->        t      = "gx_f_groot"
  external sqrt_in     :               tref ->         t ->               unit   = "gx_sqrt"
  external root_in     :               tref ->         t -> int ->        unit   = "gx_root"
  external gsqrt_in    : round_mode -> tref ->         t ->               unit   = "gx_gsqrt"
  external groot_in    : round_mode -> tref ->         t -> int ->        unit   = "gx_groot"
  external fact        :                             int ->               t      = "gx_f_fact"
  external fact_in     :               tref ->       int ->               unit   = "gx_fact"
  external gcd         :                               t -> t   ->        t      = "gx_f_gcd"
  external gcd_ex      :                               t -> t   ->        t*t*t  = "gx_f_gcd_ex"
  external cfrac       :                               t -> t   ->     t*t*t*t*t = "gx_f_cfrac"
  external gcd_in      : tref->                        t -> t   ->        unit   = "gx_gcd"
  external gcd_ex_in   : tref->tref->tref->            t -> t   ->        unit   = "gx_gcd_ex"
  external cfrac_in    : tref->tref->tref->tref->tref->t -> t   ->        unit   = "gx_cfrac_bytecode" "gx_cfrac"
  external isprime     :                               t ->           tristate   = "gx_isprime"
  external isprime_1   :                             int ->           tristate   = "gx_isprime_1"
  external sgn         :                               t ->               int    = "gx_sgn"
  external cmp         :                               t -> t   ->        int    = "gx_cmp"
  external cmp_1       :                               t -> int ->        int    = "gx_cmp_1"
  external eq          :                               t -> t   ->        bool   = "gx_eq"
  external eq_1        :                               t -> int ->        bool   = "gx_eq_1"
  external neq         :                               t -> t   ->        bool   = "gx_neq"
  external neq_1       :                               t -> int ->        bool   = "gx_neq_1"
  external inf         :                               t -> t   ->        bool   = "gx_inf"
  external inf_1       :                               t -> int ->        bool   = "gx_inf_1"
  external infeq       :                               t -> t   ->        bool   = "gx_infeq"
  external infeq_1     :                               t -> int ->        bool   = "gx_infeq_1"
  external sup         :                               t -> t   ->        bool   = "gx_sup"
  external sup_1       :                               t -> int ->        bool   = "gx_sup_1"
  external supeq       :                               t -> t   ->        bool   = "gx_supeq"
  external supeq_1     :                               t -> int ->        bool   = "gx_supeq_1"
  external of_int      :                             int ->               t      = "gx_of_int"
  external of_string   :                          string ->               t      = "gx_of_string"
  external of_int_in   :               tref ->       int ->               unit   = "gx_copy_int"
  external of_string_in:               tref ->    string ->               unit   = "gx_copy_string"
  external int_of      :                               t ->               int    = "gx_int_of"
  external string_of   :                               t ->               string = "gx_string_of"
  external bstring_of  :                               t ->               string = "gx_bstring_of"
  external hstring_of  :                               t ->               string = "gx_hstring_of"
  external ostring_of  :                               t ->               string = "gx_ostring_of"
  external nrandom     :                             int->                t      = "gx_f_nrandom"
  external zrandom     :                             int->                t      = "gx_f_zrandom"
  external nrandom1    :                             int->                t      = "gx_f_nrandom1"
  external zrandom1    :                             int->                t      = "gx_f_zrandom1"
  external nrandom_in  :               tref ->       int->                unit   = "gx_nrandom"
  external zrandom_in  :               tref ->       int->                unit   = "gx_zrandom"
  external nrandom1_in :               tref ->       int->                unit   = "gx_nrandom1"
  external zrandom1_in :               tref ->       int->                unit   = "gx_zrandom1"
  val      random_init :                             int->                unit
  external nbits       :                               t ->               int    = "gx_nbits"
  external lowbits     :                               t ->               int    = "gx_lowbits"
  external highbits    :                               t ->               int    = "gx_highbits"
  external nth_word    :                               t -> int ->        int    = "gx_nth_word"
  external nth_bit     :                               t -> int ->        bool   = "gx_nth_bit"
  external shl         :                               t -> int ->        t      = "gx_f_shl"
  external shr         :                               t -> int ->        t      = "gx_f_shr"
  external split       :                               t -> int ->        t*t    = "gx_f_split"
  external join        :                               t -> t   -> int -> t      = "gx_f_join"
  external shl_in      :               tref ->         t -> int ->        unit   = "gx_shl"
  external shr_in      :               tref ->         t -> int ->        unit   = "gx_shr"
  external split_in    :               tref -> tref -> t -> int ->        unit   = "gx_split"
  external join_in     :               tref ->         t -> t   -> int -> unit   = "gx_join"
  external dump        :                               t ->               unit   = "gx_dump"
  val toplevel_print   :                               t ->               unit
  val toplevel_print_tref:             tref ->                            unit
  exception Error of string

end

                     (* +-----------------------------+
                        |  Comparaison entre modules  |
                        +-----------------------------+ *)

module Cmp(A:Int_type)(B:Int_type) : Int_type




                           (* +----------------+
                              |  Statistiques  |
                              +----------------+ *)


module Count(A:Int_type) : sig

  type statelt = {
    mutable n:float;  (* nombre d''appels,   number of calls      *)
    mutable s:float;  (* somme des tailles,  sum of operand sizes *)
    mutable m:int     (* taille maximale,    max operand size     *)
  }

  val cadd  : statelt (* add      sub                              *)
  val cmul  : statelt (* mul      sqr                              *)
  val cquo  : statelt (* quo      modulo       quomod              *)
  val cpow  : statelt (* pow      powmod       fact                *)
  val croot : statelt (* sqrt     root                             *)
  val cgcd  : statelt (* gcd      gcd_ex       cfrac       isprime *)
  val cbin  : statelt (* shr      shl          split       join    *)
                      (* nbits    lowbits      highbits    nth_bit *)
                      (* nth_word random                           *)
  val cmisc : statelt (* abs      neg          make_ref    copy_in *)
                      (* copy_out comparaisons conversions         *)

  val clear_stats : unit -> unit (* remise  zro,   reset all counters *)
  val print_stats : unit -> unit (* affichage,       print counters     *)

  include Int_type

end

      (* +----------------------------------------------------------+
         |  Approximation des fonctions usuelles  valeurs relles  |
         +----------------------------------------------------------+ *)

module Rfuns(E:Int_type) : sig

  (* possible errors are :

     Error "0/0" when a and b are both zero
     Error "f"   when a/b is an invalid argument for function f
  *)
  exception Error of string

  (* [f a b n] returns x such that |2^n*f(a/b) - x| < 1 *)

  val arccos    : E.t -> E.t -> int -> E.t
  val arccosh   : E.t -> E.t -> int -> E.t
  val arccot    : E.t -> E.t -> int -> E.t
  val arccoth   : E.t -> E.t -> int -> E.t
  val arcsin    : E.t -> E.t -> int -> E.t
  val arcsinh   : E.t -> E.t -> int -> E.t
  val arctan    : E.t -> E.t -> int -> E.t
  val arctanh   : E.t -> E.t -> int -> E.t
  val arg       : E.t -> E.t -> int -> E.t
  val cos       : E.t -> E.t -> int -> E.t
  val cosh      : E.t -> E.t -> int -> E.t
  val cosin     : E.t -> E.t -> int -> E.t*E.t
  val cosinh    : E.t -> E.t -> int -> E.t*E.t
  val cot       : E.t -> E.t -> int -> E.t
  val coth      : E.t -> E.t -> int -> E.t
  val exp       : E.t -> E.t -> int -> E.t
  val ln        : E.t -> E.t -> int -> E.t
  val sin       : E.t -> E.t -> int -> E.t
  val sinh      : E.t -> E.t -> int -> E.t
  val tan       : E.t -> E.t -> int -> E.t
  val tanh      : E.t -> E.t -> int -> E.t

  (* [e_f a p n] returns x such that |2^n*f(a/2^p) - x| < 1 *)

  val e_arccos  : E.t -> int -> int -> E.t
  val e_arccosh : E.t -> int -> int -> E.t
  val e_arccot  : E.t -> int -> int -> E.t
  val e_arccoth : E.t -> int -> int -> E.t
  val e_arcsin  : E.t -> int -> int -> E.t
  val e_arcsinh : E.t -> int -> int -> E.t
  val e_arctan  : E.t -> int -> int -> E.t
  val e_arctanh : E.t -> int -> int -> E.t
  val e_arg     : E.t -> int -> int -> E.t
  val e_cos     : E.t -> int -> int -> E.t
  val e_cosh    : E.t -> int -> int -> E.t
  val e_cosin   : E.t -> int -> int -> E.t*E.t
  val e_cosinh  : E.t -> int -> int -> E.t*E.t
  val e_cot     : E.t -> int -> int -> E.t
  val e_coth    : E.t -> int -> int -> E.t
  val e_exp     : E.t -> int -> int -> E.t
  val e_ln      : E.t -> int -> int -> E.t
  val e_sin     : E.t -> int -> int -> E.t
  val e_sinh    : E.t -> int -> int -> E.t
  val e_tan     : E.t -> int -> int -> E.t
  val e_tanh    : E.t -> int -> int -> E.t

  (* [r_f r a b c] returns the integer approximating c*f(a/b) according
                   to the rounding mode r *)

  val r_arccos  : round_mode -> E.t -> E.t -> E.t -> E.t
  val r_arccosh : round_mode -> E.t -> E.t -> E.t -> E.t
  val r_arccot  : round_mode -> E.t -> E.t -> E.t -> E.t
  val r_arccoth : round_mode -> E.t -> E.t -> E.t -> E.t
  val r_arcsin  : round_mode -> E.t -> E.t -> E.t -> E.t
  val r_arcsinh : round_mode -> E.t -> E.t -> E.t -> E.t
  val r_arctan  : round_mode -> E.t -> E.t -> E.t -> E.t
  val r_arctanh : round_mode -> E.t -> E.t -> E.t -> E.t
  val r_arg     : round_mode -> E.t -> E.t -> E.t -> E.t
  val r_cos     : round_mode -> E.t -> E.t -> E.t -> E.t
  val r_cosh    : round_mode -> E.t -> E.t -> E.t -> E.t
  val r_cosin   : round_mode -> E.t -> E.t -> E.t -> E.t*E.t
  val r_cosinh  : round_mode -> E.t -> E.t -> E.t -> E.t*E.t
  val r_cot     : round_mode -> E.t -> E.t -> E.t -> E.t
  val r_coth    : round_mode -> E.t -> E.t -> E.t -> E.t
  val r_exp     : round_mode -> E.t -> E.t -> E.t -> E.t
  val r_ln      : round_mode -> E.t -> E.t -> E.t -> E.t
  val r_sin     : round_mode -> E.t -> E.t -> E.t -> E.t
  val r_sinh    : round_mode -> E.t -> E.t -> E.t -> E.t
  val r_tan     : round_mode -> E.t -> E.t -> E.t -> E.t
  val r_tanh    : round_mode -> E.t -> E.t -> E.t -> E.t

  (* user creation of an r_function *)
  val round : (int -> E.t) -> round_mode -> E.t -> E.t

  (* cache : the Rfuns(E) module stores in an internal cache the approximations
     of some frequently used constants. From these constants, the following
     numbers can be re-computed at practically no cost (a few additions and
     one shift) :

     ln(2)     exp(1)      arctan(1)    
     ln(3)     exp(-1)     arctan(1/2)
     ln(5)     cos(1)      arctan(1/3)
               sin(1)      arctan(1/5)

     cache_bits()  returns the sum of the bit lengths of the internal
                   constants. One bit occupies at most two memory bits.

     clear_cache() restores the internal constants to their initial
                   values, enabling the CG to reclaim memory.

  *)
  val cache_bits  : unit -> int
  val clear_cache : unit -> unit

end

                      (* +---------------------------+
                         |  Slection  l'excution  |
                         +---------------------------+ *)

module type Main_type = sig
  val main : string list -> unit
end

module Start(Main : functor(E:Int_type) -> Main_type) : sig
  val start : unit -> unit
end

